//
// hello.HC-SR501.c
//
// HC-SR501 motion detector hello-world
//    9600 baud FTDI interface
//
// Neil Gershenfeld 11/16/15
// (c) Massachusetts Institute of Technology 2015
//
// This work may be reproduced, modified, distributed,
// performed, and displayed for any purpose. Copyright is
// retained and must be preserved. The work is provided
// as is; no warranty is provided, and users accept all
// liability.
//

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#define output(directions, pin) (directions |= pin) // set port direction for output
#define input(directions,pin) (directions &= (~pin)) // set port direction for input
#define set(port, pin) (port |= pin) // set port pin
#define clear(port, pin) (port &= (~pin)) // clear port pin
#define pin_test(pins, pin) (pins & pin) // test for port pin
#define bit_test(byte, bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 102 // bit delay for 9600 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define char_delay() _delay_ms(10) // char delay

#define serial_port PORTB
#define serial_direction DDRB
#define serial_pins PINB
#define serial_pin_in (1 << PB1)
#define serial_pin_out (1 << PB2)
#define serial_interrupt (1 << PCIE0)
#define serial_interrupt_pin (1 << PCINT0)

#define sensor_pin (1 << PB3)

#define node_id '0'
#define LOW_THRESHOLD 450
#define HIGH_THRESHOLD 600


static char is_strum = 0;  // 0 for false, 1 for true

void get_char(volatile unsigned char *pins, unsigned char pin, char *rxbyte) {
  // read character into rxbyte on pins pin
  //    assumes line driver (inverts bits)

  *rxbyte = 0;
  while (pin_test(*pins,pin)); // wait for start bit

  // delay to middle of first data bit
  half_bit_delay();
  bit_delay();

  // unrolled loop to read data bits
  if pin_test(*pins,pin) *rxbyte |= (1 << 0);
  else *rxbyte |= (0 << 0);
  bit_delay();
  if pin_test(*pins,pin) *rxbyte |= (1 << 1);
  else *rxbyte |= (0 << 1);
  bit_delay();
  if pin_test(*pins,pin) *rxbyte |= (1 << 2);
  else *rxbyte |= (0 << 2);
  bit_delay();
  if pin_test(*pins,pin) *rxbyte |= (1 << 3);
  else *rxbyte |= (0 << 3);
  bit_delay();
  if pin_test(*pins,pin) *rxbyte |= (1 << 4);
  else *rxbyte |= (0 << 4);
  bit_delay();
  if pin_test(*pins,pin) *rxbyte |= (1 << 5);
  else *rxbyte |= (0 << 5);
  bit_delay();
  if pin_test(*pins,pin) *rxbyte |= (1 << 6);
  else *rxbyte |= (0 << 6);
  bit_delay();
  if pin_test(*pins,pin) *rxbyte |= (1 << 7);
  else *rxbyte |= (0 << 7);

  // wait for stop bit
  bit_delay();
  half_bit_delay();
}
void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
  // send character in txchar on port pin
  //    assumes line driver (inverts bits)

  // start bit
  clear(*port, pin);
  bit_delay();

  // unrolled loop to write data bits
  if bit_test(txchar, 0) set(*port, pin);
  else clear(*port, pin);
  bit_delay();

  if bit_test(txchar, 1) set(*port, pin);
  else clear(*port, pin);
  bit_delay();

  if bit_test(txchar, 2) set(*port, pin);
  else clear(*port, pin);
  bit_delay();

  if bit_test(txchar, 3) set(*port, pin);
  else clear(*port, pin);
  bit_delay();

  if bit_test(txchar, 4) set(*port, pin);
  else clear(*port, pin);
  bit_delay();

  if bit_test(txchar, 5) set(*port, pin);
  else clear(*port, pin);
  bit_delay();

  if bit_test(txchar, 6) set(*port, pin);
  else clear(*port, pin);
  bit_delay();

  if bit_test(txchar, 7) set(*port, pin);
  else clear(*port, pin);
  bit_delay();

  // stop bit
  set(*port, pin);
  bit_delay();

  // char delay
  bit_delay();
}
void put_string(volatile unsigned char *port, unsigned char pin, char *str) {
  // print a null-terminated string
  static int index;
  index = 0;
  do {
    put_char(port, pin, str[index]);
    ++index;
  } while (str[index] != 0);
}

ISR(PCINT0_vect) {
  // pin change interrupt handler
  static char chr;

  get_char(&serial_pins, serial_pin_in, &chr);
  if (chr == node_id) {
    // send if a strum has happened and reset
    output(serial_direction, serial_pin_out);
    put_char(&serial_port, serial_pin_out, is_strum);
    is_strum = 0;
    input(serial_direction, serial_pin_out);
  }
}

int main(void) {
  // main
  static char chrl;
  static char chrh;

  // set clock divider to /1
  CLKPR = (1 << CLKPCE);
  CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);

  // initialize serial pin
  set(serial_port, serial_pin_out);
  input(serial_direction, serial_pin_out);

  // init A/D
  ADMUX = (0 << REFS2) | (0 << REFS1) | (0 << REFS0) // Vcc ref
          | (0 << ADLAR) // right adjust
          | (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (1 << MUX0); // ADC3
  ADCSRA = (1 << ADEN) // enable
           | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // prescaler /128

  // set up pin change interrupt on input pin
  GIMSK = 0b00100000;    // turns on pin change interrupts
  PCMSK = 0b00010011;    // turn on interrupts on pins PB0, PB1, &amp;amp; PB4
//  set(GIMSK, serial_interrupt);
//  set(PCMSK0, serial_interrupt_pin);
  sei();

  // main loop
  while (1) {
    // initiate conversion
    ADCSRA |= (1 << ADSC);

    // wait for completion voltage analog -> digital
    while (ADCSRA & (1 << ADSC));
    chrl = ADCL;
    chrh = ADCH;

    int value = 256 * chrh + (unsigned char)chrl;

    if (value < LOW_THRESHOLD|| value > HIGH_THRESHOLD) {
      is_strum = 1;
    } else {
      is_strum = 0;
    }
  }
}
